(*pp ocamlrun ./version_filter -pp "camlp4o pa_extend.cmo q_MLast.cmo -loc loc -impl" *)

open Printf

open Regexp_ast
open Select_lib
open Match

let regexp = Grammar.Entry.create Pcaml.gram "regexp";;
let regexp_match_case = Grammar.Entry.create Pcaml.gram "regexp_match_case";;
let range = Grammar.Entry.create Pcaml.gram "range";;

let eval_string loc s =
#if 3.06 | 3.07pre | 3.07 | 3.07_5
  Token.eval_string s
#else
  Token.eval_string loc s

let find_named_regexp loc name =
  try Hashtbl.find named_regexps name
  with Not_found ->
    Stdpp.raise_with_loc loc
    (Failure ("Unbound regular expression " ^ name))


let handle_regexp_patt loc re =
  warnings re;
  let (num, re_name) = new_regexp () in
  let var_name = var_of_regexp re_name in
  let (re_string, named_groups, postbindings) = 
    (!lib).process_regexp loc re re_name in
  add_compiled_regexp ~anchored:true postbindings
    loc re_name num re_string named_groups;
  <:patt< $lid:var_name$ >>




let extend_common () =
EXTEND
  GLOBAL: regexp regexp_match_case range;

  Pcaml.str_item: [
    [ "RE"; name = LIDENT; "="; re = regexp -> 
	warnings re;
	Hashtbl.add named_regexps name re;
      <:str_item< declare end >> ]
  ];

  Pcaml.patt: LEVEL "simple" [
    [ "RE"; re = regexp -> handle_regexp_patt loc re
    | "/"; re = regexp; "/" -> handle_regexp_patt loc re ]
  ];

  regexp_match_case: [ 
    [ x1 = Pcaml.patt; 
      w = OPT [ "when"; e = Pcaml.expr -> e ]; "->"; 
      x2 = Pcaml.expr ->
        (loc, x1, w, x2) ]
  ];

  regexp: [
    [ r = regexp; "as"; i = LIDENT; 
      conv = 
	OPT [ ":"; s = LIDENT ->
	       (match s with
		    "int" -> `Int
		  | "float" -> `Float
		  | "option" -> `Option
		  | s -> Messages.invalid_converter loc s)
	    | ":="; e = Pcaml.expr -> `Custom e
	    | "="; e = Pcaml.expr -> `Value e  ] ->
	Bind (loc, r, i, conv) ]
  | [ r1 = regexp; "|"; r2 = regexp -> alternative loc r1 r2 ]
  | [ r1 = regexp; r2 = regexp -> Sequence (loc, r1, r2) ]

  | "postop" NONA
      [ r = regexp; "*" -> Repetition (loc, (Star, true), Closed r)
      | r = regexp; "+" -> Repetition (loc, (Plus, true), Closed r)
      | r = regexp; "?" -> Repetition (loc, (Option, true), Closed r)
      | r = regexp; "~" -> nocase r
      | r = regexp; "{"; (rng, rng_loc) = range; "}" -> 
	  if !(lib).unfold_range then repeat rng_loc (Closed r) rng
	  else Repetition (loc, (Range rng, true), Closed r) ]

  | "binop" LEFTA 
      [ r1 = regexp; "#"; r2 = regexp ->
	  let msg = " term is not a set of characters" in
	  let set1 = Regexp_ast.as_charset loc ("left" ^ msg) r1 in
	  let set2 = Regexp_ast.as_charset loc ("right" ^ msg) r2 in
	  Characters (loc, Charset.diff set1 set2) ]

  | "preop" NONA [ "!"; ident = LIDENT -> Backref (loc, ident) ]

  | "simple" NONA 
      [ "["; set = charset; "]" -> Characters (loc, set)
      | s = string -> s
      | name = LIDENT -> find_named_regexp loc name
      | "%"; name = LIDENT -> Bind_pos (loc, name)
      | "("; r = regexp; ")" -> r
      ]
  ];

  string: [
    [ s = STRING -> Regexp_ast.of_string loc (eval_string loc s) ]
  | [ c = CHAR -> Characters (loc, Charset.singleton (Token.eval_char c)) ]
  ];

  charset: [
    [ "^"; x = charset -> Charset.complement x ]
  | [ c1 = CHAR; "-"; c2 = CHAR -> 
	Charset.range (Token.eval_char c1) (Token.eval_char c2) 

    | c = CHAR -> Charset.singleton (Token.eval_char c)
    | s = STRING -> Charset.of_string (eval_string loc s)
    | name = LIDENT -> 
	Regexp_ast.as_charset loc "not a set of characters" 
	  (find_named_regexp loc name)
    | set1 = charset; set2 = charset -> Charset.union set1 set2
    ]
  ];

  range: [
    [ mini = INT;
      maxi = 
	OPT ["-"; maxi = 
	       OPT [ maxi = INT -> int_of_string maxi ] -> maxi] -> 
	  let mini = int_of_string mini in
	  (mini, maxi), loc
    | mini = INT; "+" -> (int_of_string mini, Some None), loc ]
  ];

END;;

let extend_regular () =
  extend_common ();

  EXTEND
  Pcaml.expr: LEVEL "expr1" [
    [ "match"; target = Pcaml.expr; "with"; OPT "|";
      cases = LIST1 regexp_match_case SEP "|" -> 
	output_match loc target cases
    | "try"; e = Pcaml.expr; "with"; OPT "|";
      cases = LIST1 regexp_match_case SEP "|" -> 
	output_try loc e cases
    | "function"; OPT "|"; cases = LIST1 regexp_match_case SEP "|" ->
        output_function loc cases ]
  ];

  END


let extend_revised () =
  extend_common ();

  EXTEND
  Pcaml.expr: LEVEL "top" [
    [ "match"; target = Pcaml.expr; "with"; "[";
      cases = LIST1 regexp_match_case SEP "|"; "]" -> 
	output_match loc target cases
    | "try"; e = Pcaml.expr; "with"; "[";
      cases = LIST1 regexp_match_case SEP "|"; "]" -> 
	output_try loc e cases
    | "fun"; "["; cases = LIST1 regexp_match_case SEP "|"; "]" ->
        output_function loc cases ]
  ];

  END


let () =
  init_named_regexps ();

  Pcaml.add_option "-tailrec" 
    (Arg.Set tailrec)
    " produce code that preserves tail-recursivity (default)";

  Pcaml.add_option "-direct" 
    (Arg.Clear tailrec)
    " produce code that does not try to preserve tail-recursivity";

  (match !Pcaml.syntax_name with
       "OCaml" -> extend_regular ()
     | _ -> extend_revised ())
